/*
 * created 2004/09/12
 */
package jp.sourceforge.qrcode.codec;

import java.io.IOException;

import jp.sourceforge.qrcode.codec.data.QRCodeImage;
import jp.sourceforge.qrcode.codec.data.QRCodeSymbol;
import jp.sourceforge.qrcode.codec.ecc.ReedSolomon;
import jp.sourceforge.qrcode.codec.exception.DecodingFailedException;
import jp.sourceforge.qrcode.codec.exception.InvalidDataBlockException;
import jp.sourceforge.qrcode.codec.exception.SymbolNotFoundException;
import jp.sourceforge.qrcode.codec.reader.QRCodeDataBlockReader;
import jp.sourceforge.qrcode.codec.reader.QRCodeImageReader;
import jp.sourceforge.qrcode.codec.util.DebugCanvas;
import jp.sourceforge.qrcode.codec.util.DebugCanvasAdapter;

public class QRCodeDecoder {
	QRCodeSymbol qrCodeSymbol;
	static DebugCanvas canvas = new DebugCanvasAdapter();

	public static void setCanvas(DebugCanvas canvas) {
		QRCodeDecoder.canvas = canvas;
	}

	public static DebugCanvas getCanvas() {
		return canvas;
	}

	public QRCodeDecoder() {
	}
	
/*	public String decode(QRCodeImage qrCodeImage) throws DecodingFailedException{
		canvas.println("Decoding started.");
		int[][] intImage = imageToIntArray(qrCodeImage);
		try {
			QRCodeImageReader reader = new QRCodeImageReader();
			qrCodeSymbol = reader.getQRCodeSymbol(intImage);
		} catch (SymbolNotFoundException e) {
			throw new DecodingFailedException();
		}
		canvas.println("Created QRCode symbol.");
		canvas.println("Reading symbol.");
		canvas.println("Version: " + qrCodeSymbol.getVersionReference());		
		canvas.println("Mask pattern: " + qrCodeSymbol.getMaskPatternRefererAsString());
		canvas.println("Unmasking.");
		int[] blocks = qrCodeSymbol.getBlocks();
		canvas.println("Correcting data errors.");
		int[] dataBlocks = correctDataBlocks(blocks);
		String decodedString = "";
		try {
			decodedString = getDecodedString(dataBlocks, qrCodeSymbol.getVersion());
		} catch (IllegalDataBlockException e) {
			e.printStackTrace();
			throw new DecodingFailedException();
		}
		canvas.println("Decoding finished.");
		return decodedString;
	}*/
	
	public byte[] decode(QRCodeImage qrCodeImage) throws DecodingFailedException{
		canvas.println("Decoding started.");
		int[][] intImage = imageToIntArray(qrCodeImage);
		try {
			QRCodeImageReader reader = new QRCodeImageReader();
			qrCodeSymbol = reader.getQRCodeSymbol(intImage);
		} catch (SymbolNotFoundException e) {
			throw new DecodingFailedException(e.getMessage());
		}
		canvas.println("Created QRCode symbol.");
		canvas.println("Reading symbol.");
		canvas.println("Version: " + qrCodeSymbol.getVersionReference());		
		canvas.println("Mask pattern: " + qrCodeSymbol.getMaskPatternRefererAsString());
		int[] blocks = qrCodeSymbol.getBlocks();
		canvas.println("Correcting data errors.");
		int[] dataBlocks = correctDataBlocks(blocks);
		try {
			byte[] decodedByteArray = 
				getDecodedByteArray(dataBlocks, qrCodeSymbol.getVersion());
			canvas.println("Decoding finished.");
			return decodedByteArray;
		} catch (InvalidDataBlockException e) {
			throw new DecodingFailedException(e.getMessage());
		}
	}	
	
	
	int[][] imageToIntArray(QRCodeImage image) {
		int width = image.getWidth();
		int height = image.getHeight();
		int[][] intImage = new int[width][height];
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				intImage[x][y] = image.getPixel(x,y);
			}
		}
		return intImage;
	}
	
	int[] correctDataBlocks(int[] blocks) {
		int numErrors = 0;
		int dataCapacity = qrCodeSymbol.getDataCapacity();
		int[] dataBlocks = new int[dataCapacity];
		int numErrorCollectionCode = qrCodeSymbol.getNumErrorCollectionCode();
		int numRSBlocks = qrCodeSymbol.getNumRSBlocks();
		int eccPerRSBlock = numErrorCollectionCode / numRSBlocks;
		if (numRSBlocks == 1) {
			ReedSolomon corrector = new ReedSolomon(blocks);
			corrector.correct();
			numErrors += corrector.getNumCorrectedErrors();
			if (numErrors > 0)
				canvas.println(String.valueOf(numErrors) + " data errors corrected.");
			else
				canvas.println("No errors found.");				
			return blocks;
		}
		else  { //we have to interleave data blocks because symbol has 2 or more RS blocks
			int numLongerRSBlocks = dataCapacity % numRSBlocks;
			if (numLongerRSBlocks == 0) { //symbol has only 1 type of RS block
				int lengthRSBlock = dataCapacity / numRSBlocks;
				int[][] RSBlocks = new int[numRSBlocks][lengthRSBlock];
				//obtain RS blocks
				for (int i = 0; i < numRSBlocks; i++) {
					for (int j = 0; j < lengthRSBlock; j++) {
						RSBlocks[i][j] = blocks[j * numRSBlocks + i];
					}
					ReedSolomon corrector = new ReedSolomon(RSBlocks[i]);
					corrector.correct();
					numErrors += corrector.getNumCorrectedErrors();
				}
				//obtain only data part
				int p = 0;
				for (int i = 0; i < numRSBlocks; i++) {
					for (int j = 0; j < lengthRSBlock - eccPerRSBlock; j++) {
						dataBlocks[p++] = RSBlocks[i][j];
					}
				}
			}
			else { //symbol has 2 types of RS blocks
				int lengthShorterRSBlock = dataCapacity / numRSBlocks;
				int lengthLongerRSBlock = dataCapacity / numRSBlocks + 1;
				int numShorterRSBlocks = numRSBlocks - numLongerRSBlocks;
				int[][] shorterRSBlocks = new int[numShorterRSBlocks][lengthShorterRSBlock];
				int[][] longerRSBlocks = new int[numLongerRSBlocks][lengthLongerRSBlock];
				for (int i = 0; i < numRSBlocks; i++) {
					if (i < numShorterRSBlocks) { //get shorter RS Block(s)
						int mod = 0;
						for (int j = 0; j < lengthShorterRSBlock; j++) {
							if (j == lengthShorterRSBlock - eccPerRSBlock) mod = numLongerRSBlocks;
							shorterRSBlocks[i][j] = blocks[j * numRSBlocks + i + mod];
						}
						ReedSolomon corrector = new ReedSolomon(shorterRSBlocks[i]);
						corrector.correct();
						numErrors += corrector.getNumCorrectedErrors();
					}
					else { 	//get longer RS Blocks
						int mod = 0;
						for (int j = 0; j < lengthLongerRSBlock; j++) {
							if (j == lengthShorterRSBlock - eccPerRSBlock) mod = numShorterRSBlocks;
							longerRSBlocks[i - numShorterRSBlocks][j] = blocks[j * numRSBlocks + i - mod];
						}
						
						ReedSolomon corrector = new ReedSolomon(longerRSBlocks[i - numShorterRSBlocks]);
						corrector.correct();
						numErrors += corrector.getNumCorrectedErrors();
					}
				}
				int p = 0;
				for (int i = 0; i < numRSBlocks; i++) {
					if (i < numShorterRSBlocks) {
						for (int j = 0; j < lengthShorterRSBlock - eccPerRSBlock; j++) {
							dataBlocks[p++] = shorterRSBlocks[i][j];
						}
					}
					else {
						for (int j = 0; j < lengthLongerRSBlock - eccPerRSBlock; j++) {
							dataBlocks[p++] = longerRSBlocks[i - numShorterRSBlocks][j];
						}
					}
				}
			}
			if (numErrors > 0)
				canvas.println(String.valueOf(numErrors) + " data errors corrected.");
			return dataBlocks;
		}
	}
	
	byte[] getDecodedByteArray(int[] blocks, int version) throws InvalidDataBlockException {
		byte[] byteArray;
		QRCodeDataBlockReader reader = new QRCodeDataBlockReader(blocks, version);
		try {
			byteArray = reader.getDataByte();
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new InvalidDataBlockException(e.getMessage());
		} catch (IOException e) {
			throw new InvalidDataBlockException(e.getMessage());
		}
		return byteArray;
	}

	String getDecodedString(int[] blocks, int version) throws InvalidDataBlockException {
		String dataString = null;
		QRCodeDataBlockReader reader = new QRCodeDataBlockReader(blocks, version);
		try {
			dataString = reader.getDataString();
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new InvalidDataBlockException(e.getMessage());
		}
		return dataString;
	}
}

